public class Expression {
    private static enum Action { Push, Pop }
    private static final Action[][] ACT;
    
    static {
        // ACT: an action table. based on the input token and the top symbol
        //      of the stack, what action needs to be taken.
        // Push: push the current input token to the stack
        // Pop: pop the operator from the stack and perform the operation
        // Undefined elements (null) mean an error

        ACT = new Action[128][128]; //[stack top][input token]
        
        //TODO: add table entries for unary minus '~' operator

        //stack top is '#'
        for(char c : "(+-*/".toCharArray())     ACT['#'][c] = Action.Push;
        for(char c : "$".toCharArray())         ACT['#'][c] = Action.Pop;

        //stack top is '('
        for(char c : "(+-*/".toCharArray())     ACT['('][c] = Action.Push;
        for(char c : ")".toCharArray())         ACT['('][c] = Action.Pop;

        //stack top is '+'
        for(char c : "(*/".toCharArray())       ACT['+'][c] = Action.Push;
        for(char c : "+-)$".toCharArray())      ACT['+'][c] = Action.Pop;

        //stack top is '-'
        for(char c : "(*/".toCharArray())       ACT['-'][c] = Action.Push;
        for(char c : "+-)$".toCharArray())      ACT['-'][c] = Action.Pop;

        //stack top is '*'
        for(char c : "(".toCharArray())         ACT['*'][c] = Action.Push;
        for(char c : "+-*/)$".toCharArray())    ACT['*'][c] = Action.Pop;

        //stack top is '/'
        for(char c : "(".toCharArray())         ACT['/'][c] = Action.Push;
        for(char c : "+-*/)$".toCharArray())    ACT['/'][c] = Action.Pop;
    }
    
    public static double evalExpr(String expr) {
        //TODO: handle unary minus operator
        Scanner scan = new Scanner(expr);
        Stack<Double> num    = new StackByArray<Double>();
        Stack<Character> opr = new StackByArray<Character>();
        
        opr.push('#');
        for(String tok : scan) {
            char cur = tok.charAt(0);
            double n1, n2;  //operands

            if(Scanner.isDigit(cur)) {
                num.push(Double.parseDouble(tok));
            }
            else {
                Action act;
                loop: while((act = ACT[opr.top()][cur]) == Action.Pop) {
                    char op = opr.pop();
                    switch(op) {
                    case '+': n2 = num.pop(); n1 = num.pop(); num.push(n1 + n2); break;
                    case '-': n2 = num.pop(); n1 = num.pop(); num.push(n1 - n2); break;
                    case '*': n2 = num.pop(); n1 = num.pop(); num.push(n1 * n2); break;
                    case '/': n2 = num.pop(); n1 = num.pop(); num.push(n1 / n2); break;
                    case '#':
                    case '(': break loop; //cur cancels op
                    default:
                        throw new UnsupportedOperationException("Error: " + tok);
                    }
                }
                if(act == Action.Push)
                    opr.push(cur);
                else if(act == null)
                    throw new IllegalStateException("Syntax error: " + tok);
            }
        }
        if(opr.size() != 0 || num.size() != 1)
            throw new IllegalStateException("Syntax error");
        return num.pop();
    }
        
    public static String infixToPostfix(String expr) {
        //TODO: implement this method
    }
    
    public static double evalPostfixExpr(String expr) {
        //TODO: implement this method
    }    
}
